/* http.c -- Dune Webserver
 * Copyright (C) 1998 Baris Mestanogullari <mestanba@martin.luther.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <time.h>
#include <pwd.h>

#include "http.h"
#include "config.h"

int goodFileName(char *url) {
  if(strstr(url, "//"))
    return 0;
  if(strstr(url, ".." ))
    return 0;
  return 1;
}

char *getmt(struct tm *t) { /* Day, 00 Mon 0000 00:00:00 GMT */
  static char mtime[44];
  char *day[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
  char *month[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
                    "Sep", "Oct", "Nov", "Dec" };
  sprintf(mtime, "%s, %02i %s %d %02i:%02i:%02i GMT", day[t->tm_wday], 
          t->tm_mday, month[t->tm_mon], t->tm_year+1900, t->tm_hour,
          t->tm_min, t->tm_sec);
  return mtime;
}

char *currentTime() {
  time_t t;
  t=time(NULL);
  return getmt(gmtime(&t));
}

char *contentType(char *url) {
  int counter=0, counter2=strlen(url);
  if(strstr(url, ".")) {
    while(url[counter2]!='.') counter2--;
    while(url[counter2]!='\0') {
      buffer[counter]=url[counter2];
      counter++;
      counter2++;
    }
    buffer[counter]='\0';
    for(counter=0; types[counter].extension; counter++ )
      if(!strcmp(buffer, types[counter].extension))
	return types[counter].type;
    return "text/plain";
  }
  return "text/plain";
}

int checkFile(char *pathnm) {
  FILE *ftest;
  DIR *dtest;
  ftest=fopen(pathnm, "r");
  if(!ftest) { 
    if(errno==ENOENT)  /* Does not exist */
      return E_NOENT;
    if(errno==EACCES)  /* No permissions */
      return E_ACCES;
  }
  dtest=opendir(pathnm);
  
  if(dtest)  /* It is a directory */
    return IS_DIR;

  /* Now we know it exists, it is not a directory and it has the permissions */
  return IS_FILE;
}

char *multiUser(char *fileName, char *pageRequested) {
/* Gets 'pageRequested' in the form /~baris/blah.html and changes that to 
 * /home/baris/blah.html. (Or whatever the homedir is)
 */
  char userName[16], theRest[1024];
  struct passwd *pwd;
  int counter=0, counter2=0;
  strcpy(fileName, pageRequested);

  fileName+=2;
  if(strstr(fileName, "/")) {
    for(counter=0; fileName[counter]!='/'; counter++) {
      if(counter>14)
	return pageRequested;
      userName[counter]=fileName[counter];
    }
    userName[counter]='\0';
    for(; fileName[counter]!='\0'; counter++) {
      theRest[counter2]=fileName[counter];
      counter2++;
      if(counter2>1022)
	return pageRequested;
    }
    theRest[counter2++]='\0';
  } else {
    for(counter=0; fileName[counter]!='\0'; counter++) {
      userName[counter]=fileName[counter];
    }
    userName[counter]='\0';
    theRest[0]=0;
  }
  do {
    if((pwd=getpwent())==NULL) { return pageRequested; }
  } while(strcmp(pwd->pw_name, userName));

  bzero(fileName, sizeof(fileName));
  sprintf(fileName, "%s/%s/", pwd->pw_dir, USER_DIR);
  if(theRest[0]!='\0' && strcmp(theRest, "/")) {
    strcat(fileName, theRest);
  }
  endpwent();
  return fileName;
}

void processRequest(char *received, char *remoteAddress) {
/* Big and ugly function */
  char *pageRequested;
  char command[5], tmp[256], dirname[256];
  char fileName[1024];
  int flag=0, head=0, numeric=0, user=0, counter=0, counter2=0;

  pageRequested=(char *)malloc(256);
  bzero(pageRequested,sizeof(pageRequested));

  while(received[counter]!=' ') {
    if(counter>4 || received[counter]=='\0' || received[counter]=='\r' || received[counter]=='\n') {
      sendBadRequest(0);
#ifdef LOG_CONNECTIONS
      received[strlen(received)-1]='\0';
      LOG(remoteAddress, received, 400);
#endif
      return;
    }
    command[counter]=received[counter++];
  }
  command[counter]='\0';

  counter++;
  while(received[counter]!=' ' && received[counter]!='\0' && received[counter]!='\r' && received[counter]!='\n') {
    if(counter2>254) {
      sendBadRequest(!strcmp(command, "HEAD"));
#ifdef LOG_CONNECTIONS
      received[strlen(received)-1]='\0';
      LOG(remoteAddress, received, 400);
#endif
      return;
    }
    pageRequested[counter2++]=received[counter++];
  }
  pageRequested[counter2]='\0';

  if(pageRequested[0]!='/') {
    sendBadRequest(!strcmp(command, "HEAD"));
#ifdef LOG_CONNECTIONS
    received[strlen(received)-1]='\0';
    LOG(remoteAddress, received, 400);
#endif
    return;
  }
  if (!strcmp(command,"GET") || !strcmp(command,"HEAD")) {
    if(!strcmp(command,"GET")) head=0;
    else head=1;
    if(pageRequested[1]=='~')
      user=1;

    if(!goodFileName(pageRequested)) {
      sendBadRequest(head);
      numeric=400;
    }

    else if(pageRequested[strlen(pageRequested)-1]=='/') {

      if(!user) sprintf(tmp, "%s%s", HTMLDIR, pageRequested);
      else { strcpy(tmp, multiUser(fileName, pageRequested)); user=0; }

      if(checkFile(tmp)==E_ACCES) {
	sendForbidden(pageRequested, head);
	numeric=403;
      }
      else if(checkFile(tmp)!=IS_DIR) {
	sendNotFound(pageRequested, head);
	numeric=404;
      }
      else if(checkFile(tmp)==IS_DIR) {
	strcpy(dirname, tmp);
	strcat(tmp, "index.html");
	if(checkFile(tmp)==IS_FILE) {
          if(head==0) {
	    sendFile(tmp, 1);
	    numeric=200;
	  }
          else {
	    sendContentInfo(tmp);
	    numeric=200;
	  }
        }
	else if(checkFile(tmp)==E_ACCES) {
	  sendForbidden(pageRequested, head);
	  numeric=403;
	}
	else if(checkFile(tmp)==E_NOENT) {
          sendIndex(dirname, pageRequested, head);
	  numeric=200;
	}
      }
      else if(checkFile(tmp)==E_ACCES) {
	sendForbidden(pageRequested, head);
	numeric=403;
      }
    }
    else {
      if(!user) sprintf(tmp, "%s%s", HTMLDIR, pageRequested);
      else { strcpy(tmp, multiUser(fileName, pageRequested)); user=0; }

      if(checkFile(tmp)==E_NOENT) {
	sendNotFound(pageRequested, head);
	numeric=404;
      }
      else if(checkFile(tmp)==E_ACCES) {
	sendForbidden(pageRequested, head);
	numeric=403;
      }
      else if(checkFile(tmp)==IS_DIR) {
	sendMovedPermanently(pageRequested, head);
	numeric=301;
      }
      else if(checkFile(tmp)==IS_FILE && !flag) {
        if(head==0) {
	  sendFile(tmp, 1);
	  numeric=200;
	}
        else {
	  sendContentInfo(tmp);
	  numeric=200;
	}
      }
    }
  } else {
    sendNotImplemented();
    numeric=501;
  }
#ifdef LOG_CONNECTIONS
    received[strlen(received)-1]='\0';
    LOG(remoteAddress, received, numeric);
#endif
  free(pageRequested);
}
